# -*- coding: utf-8 -*-
from .interface import LeafFactor

__all__ = [
    'Pricing',
    'Fundamentals',
    'Factor',
]


class _Factor(LeafFactor):
    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return 'Factor({!r})'.format(self.name)

    def get_series(self, data_source, order_book_id, end_date, length):
        return data_source.get_factor(order_book_id, self.name, end_date, length)

    def __eq__(self, other):
        return isinstance(other, _Factor) and other.name == self.name


_factors = {}


def Factor(name):
    """
    基本叶子节点因子

    Parameters
    ----------
    name: str
        因子名,可以为:
        * 行情因子(默认后复权, 添加_unadjusted为不复权), open, close, high, low, volume, total_turnover
        * 财务因子, num_trades, pb_ratio, pe_ratio, market_cap, ..., etc.
        * 私有因子(private.因子名)

    Notes
    -----
    财务因子名可以通过`rqdatac.get_all_factor_names()`获取
    """
    try:
        return _factors[name]
    except KeyError:
        f = _Factor(name)
        _factors[name] = f
        return f


class Pricing:
    open = Factor('open')
    close = Factor('close')
    high = Factor('high')
    low = Factor('low')
    volume = Factor('volume')
    total_turnover = Factor('total_turnover')


class Fundamentals:
    num_trades = Factor('num_trades')
    pe_ratio = Factor("pe_ratio")
    pb_ratio = Factor("pb_ratio")
    market_cap = Factor("market_cap")
    a_share_market_val = Factor("a_share_market_val")
    a_share_market_val_2 = Factor("a_share_market_val_2")
    ev = Factor("ev")
    ev_2 = Factor("ev_2")
    ev_to_ebit = Factor("ev_to_ebit")
    dividend_yield = Factor("dividend_yield")
    pe_ratio_1 = Factor("pe_ratio_1")
    peg_ratio = Factor("peg_ratio")
    ps_ratio = Factor("ps_ratio")
    ebitda = Factor("ebitda")
    invested_capital = Factor("invested_capital")
    fcff = Factor("fcff")
    fcfe = Factor("fcfe")
    depreciation_and_amortization = Factor("depreciation_and_amortization")
    adjusted_return_on_equity_average = Factor("adjusted_return_on_equity_average")
    annual_return_on_equity = Factor("annual_return_on_equity")
    annual_return_on_asset = Factor("annual_return_on_asset")
    annual_return_on_asset_net_profit = Factor("annual_return_on_asset_net_profit")
    debt_to_equity_ratio = Factor("debt_to_equity_ratio")
    equity_to_debt_ratio = Factor("equity_to_debt_ratio")
    equity_to_interest_bearing_debt = Factor("equity_to_interest_bearing_debt")
    inc_revenue = Factor("inc_revenue")
    du_asset_turnover_ratio = Factor("du_asset_turnover_ratio")
    earnings_per_share = Factor("earnings_per_share")
    fully_diluted_earnings_per_share = Factor("fully_diluted_earnings_per_share")
    diluted_earnings_per_share = Factor("diluted_earnings_per_share")
    adjusted_earnings_per_share = Factor("adjusted_earnings_per_share")
    adjusted_fully_diluted_earnings_per_share = Factor("adjusted_fully_diluted_earnings_per_share")
    adjusted_diluted_earnings_per_share = Factor("adjusted_diluted_earnings_per_share")
    book_value_per_share = Factor("book_value_per_share")
    operating_cash_flow_per_share = Factor("operating_cash_flow_per_share")
    operating_total_revenue_per_share = Factor("operating_total_revenue_per_share")
    operating_revenue_per_share = Factor("operating_revenue_per_share")
    capital_reserve_per_share = Factor("capital_reserve_per_share")
    earned_reserve_per_share = Factor("earned_reserve_per_share")
    undistributed_profit_per_share = Factor("undistributed_profit_per_share")
    retained_earnings_per_share = Factor("retained_earnings_per_share")
    cash_flow_from_operations_per_share = Factor("cash_flow_from_operations_per_share")
    ebit_per_share = Factor("ebit_per_share")
    free_cash_flow_company_per_share = Factor("free_cash_flow_company_per_share")
    free_cash_flow_equity_per_share = Factor("free_cash_flow_equity_per_share")
    dividend_per_share = Factor("dividend_per_share")
    return_on_equity = Factor("return_on_equity")
    return_on_equity_weighted_average = Factor("return_on_equity_weighted_average")
    return_on_equity_diluted = Factor("return_on_equity_diluted")
    adjusted_return_on_equity_weighted_average = Factor("adjusted_return_on_equity_weighted_average")
    adjusted_return_on_equity_diluted = Factor("adjusted_return_on_equity_diluted")
    return_on_asset = Factor("return_on_asset")
    return_on_asset_net_profit = Factor("return_on_asset_net_profit")
    return_on_invested_capital = Factor("return_on_invested_capital")
    net_profit_margin = Factor("net_profit_margin")
    gross_profit_margin = Factor("gross_profit_margin")
    cost_to_sales = Factor("cost_to_sales")
    net_profit_to_revenue = Factor("net_profit_to_revenue")
    profit_from_operation_to_revenue = Factor("profit_from_operation_to_revenue")
    ebit_to_revenue = Factor("ebit_to_revenue")
    expense_to_revenue = Factor("expense_to_revenue")
    operating_profit_to_profit_before_tax = Factor("operating_profit_to_profit_before_tax")
    invesment_profit_to_profit_before_tax = Factor("invesment_profit_to_profit_before_tax")
    non_operating_profit_to_profit_before_tax = Factor("non_operating_profit_to_profit_before_tax")
    income_tax_to_profit_before_tax = Factor("income_tax_to_profit_before_tax")
    adjusted_profit_to_total_profit = Factor("adjusted_profit_to_total_profit")
    debt_to_asset_ratio = Factor("debt_to_asset_ratio")
    equity_multiplier = Factor("equity_multiplier")
    current_asset_to_total_asset = Factor("current_asset_to_total_asset")
    non_current_asset_to_total_asset = Factor("non_current_asset_to_total_asset")
    interest_bearing_debt_to_capital = Factor("interest_bearing_debt_to_capital")
    current_debt_to_total_debt = Factor("current_debt_to_total_debt")
    non_current_debt_to_total_debt = Factor("non_current_debt_to_total_debt")
    current_ratio = Factor("current_ratio")
    quick_ratio = Factor("quick_ratio")
    super_quick_ratio = Factor("super_quick_ratio")
    ebit_to_debt = Factor("ebit_to_debt")
    ocf_to_debt = Factor("ocf_to_debt")
    ocf_to_interest_bearing_debt = Factor("ocf_to_interest_bearing_debt")
    ocf_to_current_ratio = Factor("ocf_to_current_ratio")
    ocf_to_net_debt = Factor("ocf_to_net_debt")
    time_interest_earned_ratio = Factor("time_interest_earned_ratio")
    account_payable_turnover_rate = Factor("account_payable_turnover_rate")
    account_payable_turnover_days = Factor("account_payable_turnover_days")
    account_receivable_turnover_days = Factor("account_receivable_turnover_days")
    inventory_turnover = Factor("inventory_turnover")
    account_receivable_turnover_rate = Factor("account_receivable_turnover_rate")
    current_asset_turnover = Factor("current_asset_turnover")
    fixed_asset_turnover = Factor("fixed_asset_turnover")
    total_asset_turnover = Factor("total_asset_turnover")
    inc_earnings_per_share = Factor("inc_earnings_per_share")
    inc_diluted_earnings_per_share = Factor("inc_diluted_earnings_per_share")
    inc_operating_revenue = Factor("inc_operating_revenue")
    inc_gross_profit = Factor("inc_gross_profit")
    inc_profit_before_tax = Factor("inc_profit_before_tax")
    inc_net_profit = Factor("inc_net_profit")
    inc_adjusted_net_profit = Factor("inc_adjusted_net_profit")
    inc_cash_from_operations = Factor("inc_cash_from_operations")
    inc_return_on_equity = Factor("inc_return_on_equity")
    inc_book_per_share = Factor("inc_book_per_share")
    inc_total_asset = Factor("inc_total_asset")
    du_return_on_equity = Factor("du_return_on_equity")
    du_equity_multiplier = Factor("du_equity_multiplier")
    du_profit_margin = Factor("du_profit_margin")
    du_return_on_sales = Factor("du_return_on_sales")
    non_recurring_profit_and_loss = Factor("non_recurring_profit_and_loss")
    adjusted_net_profit = Factor("adjusted_net_profit")
    ebit = Factor("ebit")
    working_capital = Factor("working_capital")
    net_working_capital = Factor("net_working_capital")
    retained_earnings = Factor("retained_earnings")
    interest_bearing_debt = Factor("interest_bearing_debt")
    net_debt = Factor("net_debt")
    non_interest_bearing_current_debt = Factor("non_interest_bearing_current_debt")
    non_interest_bearing_non_current_debt = Factor("non_interest_bearing_non_current_debt")
    non_operating_net_profit = Factor("non_operating_net_profit")
    other_operating_income = Factor("other_operating_income")
    inventory_shrinkage = Factor("inventory_shrinkage")
    operating_expense = Factor("operating_expense")
    investment_income = Factor("investment_income")
    subsidy_income = Factor("subsidy_income")
    pnl_adjustment = Factor("pnl_adjustment")
    profit_from_ma = Factor("profit_from_ma")
    retained_profit_at_beginning = Factor("retained_profit_at_beginning")
    profit_available_for_distribution = Factor("profit_available_for_distribution")
    statutory_welfare_reserve = Factor("statutory_welfare_reserve")
    staff_incentive_welfare_reserve = Factor("staff_incentive_welfare_reserve")
    profit_available_for_owner_distribution = Factor("profit_available_for_owner_distribution")
    preferred_stock_dividends = Factor("preferred_stock_dividends")
    ordinary_stock_dividends = Factor("ordinary_stock_dividends")
    revenue = Factor("revenue")
    operating_revenue = Factor("operating_revenue")
    sales_discount = Factor("sales_discount")
    total_expense = Factor("total_expense")
    cost_of_goods_sold = Factor("cost_of_goods_sold")
    sales_tax = Factor("sales_tax")
    gross_profit = Factor("gross_profit")
    selling_expense = Factor("selling_expense")
    ga_expense = Factor("ga_expense")
    financing_expense = Factor("financing_expense")
    exchange_gains_or_losses = Factor("exchange_gains_or_losses")
    asset_depreciation = Factor("asset_depreciation")
    profit_from_operation = Factor("profit_from_operation")
    non_operating_revenue = Factor("non_operating_revenue")
    non_operating_expense = Factor("non_operating_expense")
    disposal_loss_on_asset = Factor("disposal_loss_on_asset")
    profit_before_tax = Factor("profit_before_tax")
    income_tax = Factor("income_tax")
    unrealised_investment_losses = Factor("unrealised_investment_losses")
    income_tax_refund = Factor("income_tax_refund")
    net_profit = Factor("net_profit")
    net_profit_parent_company = Factor("net_profit_parent_company")
    enterprise_expansion_reserve = Factor("enterprise_expansion_reserve")
    other_surplus_reserve = Factor("other_surplus_reserve")
    basic_earnings_per_share = Factor("basic_earnings_per_share")
    other_income = Factor("other_income")
    total_income = Factor("total_income")
    total_income_parent_company = Factor("total_income_parent_company")
    total_income_minority = Factor("total_income_minority")
    cash = Factor("cash")
    financial_asset_held_for_trading = Factor("financial_asset_held_for_trading")
    cash_equivalent = Factor("cash_equivalent")
    bill_receivable = Factor("bill_receivable")
    devidend_receivable = Factor("devidend_receivable")
    interest_receivable = Factor("interest_receivable")
    bad_debt_reserve = Factor("bad_debt_reserve")
    prepayment = Factor("prepayment")
    subsidy_receivable = Factor("subsidy_receivable")
    prepaid_tax = Factor("prepaid_tax")
    inventory = Factor("inventory")
    deferred_expense = Factor("deferred_expense")
    contract_work = Factor("contract_work")
    long_term_debt_due_one_year = Factor("long_term_debt_due_one_year")
    non_current_debt_due_one_year = Factor("non_current_debt_due_one_year")
    other_current_assets = Factor("other_current_assets")
    current_assets = Factor("current_assets")
    financial_asset_available_for_sale = Factor("financial_asset_available_for_sale")
    financial_asset_hold_to_maturity = Factor("financial_asset_hold_to_maturity")
    real_estate_investment = Factor("real_estate_investment")
    long_term_equity_investment = Factor("long_term_equity_investment")
    long_term_receivables = Factor("long_term_receivables")
    net_long_term_equity_investment = Factor("net_long_term_equity_investment")
    engineer_material = Factor("engineer_material")
    construction_in_progress = Factor("construction_in_progress")
    fixed_asset_to_be_disposed = Factor("fixed_asset_to_be_disposed")
    capitalized_biological_assets = Factor("capitalized_biological_assets")
    oil_and_gas_assets = Factor("oil_and_gas_assets")
    intangible_assets = Factor("intangible_assets")
    impairment_intangible_assets = Factor("impairment_intangible_assets")
    goodwill = Factor("goodwill")
    long_term_deferred_expenses = Factor("long_term_deferred_expenses")
    deferred_income_tax_assets = Factor("deferred_income_tax_assets")
    other_non_current_assets = Factor("other_non_current_assets")
    non_current_assets = Factor("non_current_assets")
    total_assets = Factor("total_assets")
    short_term_loans = Factor("short_term_loans")
    financial_liabilities = Factor("financial_liabilities")
    notes_payable = Factor("notes_payable")
    accts_payable = Factor("accts_payable")
    advance_from_customers = Factor("advance_from_customers")
    proxy_sale_revenue = Factor("proxy_sale_revenue")
    payroll_payable = Factor("payroll_payable")
    dividend_payable = Factor("dividend_payable")
    tax_payable = Factor("tax_payable")
    interest_payable = Factor("interest_payable")
    other_fees_payable = Factor("other_fees_payable")
    other_payable = Factor("other_payable")
    short_term_debt = Factor("short_term_debt")
    accrued_expense = Factor("accrued_expense")
    estimated_liabilities = Factor("estimated_liabilities")
    deferred_income = Factor("deferred_income")
    other_current_liabilities = Factor("other_current_liabilities")
    current_liabilities = Factor("current_liabilities")
    long_term_loans = Factor("long_term_loans")
    bond_payable = Factor("bond_payable")
    long_term_payable = Factor("long_term_payable")
    grants_received = Factor("grants_received")
    deferred_income_tax_liabilities = Factor("deferred_income_tax_liabilities")
    other_non_current_liabilities = Factor("other_non_current_liabilities")
    non_current_liabilities = Factor("non_current_liabilities")
    total_liabilities = Factor("total_liabilities")
    paid_in_capital = Factor("paid_in_capital")
    investment_refund = Factor("investment_refund")
    capital_reserve = Factor("capital_reserve")
    surplus_reserve = Factor("surplus_reserve")
    unrealised_investment_loss = Factor("unrealised_investment_loss")
    undistributed_profit = Factor("undistributed_profit")
    equity_parent_company = Factor("equity_parent_company")
    total_equity = Factor("total_equity")
    minority_interest = Factor("minority_interest")
    total_equity_and_liabilities = Factor("total_equity_and_liabilities")
    provision = Factor("provision")
    deferred_revenue = Factor("deferred_revenue")
    current_investment = Factor("current_investment")
    current_investment_reserve = Factor("current_investment_reserve")
    net_current_investment = Factor("net_current_investment")
    accts_receivable = Factor("accts_receivable")
    other_accts_receivable = Factor("other_accts_receivable")
    net_accts_receivable = Factor("net_accts_receivable")
    other_receivables = Factor("other_receivables")
    inventory_depreciation_reserve = Factor("inventory_depreciation_reserve")
    net_inventory = Factor("net_inventory")
    long_term_debt_investment = Factor("long_term_debt_investment")
    other_long_term_investment = Factor("other_long_term_investment")
    long_term_investment = Factor("long_term_investment")
    provision_long_term_investment = Factor("provision_long_term_investment")
    net_long_term_debt_investment = Factor("net_long_term_debt_investment")
    net_long_term_investment = Factor("net_long_term_investment")
    cost_fixed_assets = Factor("cost_fixed_assets")
    accumulated_depreciation = Factor("accumulated_depreciation")
    net_val_fixed_assets = Factor("net_val_fixed_assets")
    depreciation_reserve = Factor("depreciation_reserve")
    net_fixed_assets = Factor("net_fixed_assets")
    total_fixed_assets = Factor("total_fixed_assets")
    deferred_charges = Factor("deferred_charges")
    other_long_term_assets = Factor("other_long_term_assets")
    total_intangible_and_other_assets = Factor("total_intangible_and_other_assets")
    walfare_payable = Factor("walfare_payable")
    internal_accts_payable = Factor("internal_accts_payable")
    long_term_liabilities_due_one_year = Factor("long_term_liabilities_due_one_year")
    housing_revolving_funds = Factor("housing_revolving_funds")
    other_long_term_liabilities = Factor("other_long_term_liabilities")
    long_term_liabilities = Factor("long_term_liabilities")
    statutory_reserve = Factor("statutory_reserve")
    rental_cash = Factor("rental_cash")
    refunds_of_vat = Factor("refunds_of_vat")
    refunds_of_other_taxes = Factor("refunds_of_other_taxes")
    cash_paid_for_rental = Factor("cash_paid_for_rental")
    cash_received_from_disposal_of_investment = Factor("cash_received_from_disposal_of_investment")
    cash_received_from_dividend = Factor("cash_received_from_dividend")
    cash_received_from_interest = Factor("cash_received_from_interest")
    cash_paid_for_asset = Factor("cash_paid_for_asset")
    cash_received_from_equity_investors = Factor("cash_received_from_equity_investors")
    cash_received_from_debt_investors = Factor("cash_received_from_debt_investors")
    cash_received_from_sales_of_goods = Factor("cash_received_from_sales_of_goods")
    refunds_of_taxes = Factor("refunds_of_taxes")
    cash_from_other_operating_activities = Factor("cash_from_other_operating_activities")
    cash_from_operating_activities = Factor("cash_from_operating_activities")
    cash_paid_for_goods_and_services = Factor("cash_paid_for_goods_and_services")
    cash_paid_for_employee = Factor("cash_paid_for_employee")
    cash_paid_for_taxes = Factor("cash_paid_for_taxes")
    cash_paid_for_other_operation_activities = Factor("cash_paid_for_other_operation_activities")
    cash_paid_for_operation_activities = Factor("cash_paid_for_operation_activities")
    cash_flow_from_operating_activities = Factor("cash_flow_from_operating_activities")
    cash_received_from_disposal_of_asset = Factor("cash_received_from_disposal_of_asset")
    cash_received_from_other_investment_activities = Factor("cash_received_from_other_investment_activities")
    cash_received_from_investment_activities = Factor("cash_received_from_investment_activities")
    cash_paid_to_acquire_investment = Factor("cash_paid_to_acquire_investment")
    cash_paid_for_other_investment_activities = Factor("cash_paid_for_other_investment_activities")
    cash_paid_for_investment_activities = Factor("cash_paid_for_investment_activities")
    cash_flow_from_investing_activities = Factor("cash_flow_from_investing_activities")
    cash_received_from_investors = Factor("cash_received_from_investors")
    cash_received_from_financial_institution_borrows = Factor("cash_received_from_financial_institution_borrows")
    cash_received_from_other_financing_activities = Factor("cash_received_from_other_financing_activities")
    cash_received_from_financing_activities = Factor("cash_received_from_financing_activities")
    cash_paid_for_debt = Factor("cash_paid_for_debt")
    cash_paid_for_dividend_and_interest = Factor("cash_paid_for_dividend_and_interest")
    cash_paid_for_other_financing_activities = Factor("cash_paid_for_other_financing_activities")
    cash_paid_to_financing_activities = Factor("cash_paid_to_financing_activities")
    cash_flow_from_financing_activities = Factor("cash_flow_from_financing_activities")
    capital = Factor('capital')
